/* main aglo library routine */
/* $Id: aglo.c,v 1.0 1992/10/24 03:49:38 coleman Exp $ */

#include <aglo.h>
#include <aglo_graph_io.h>
#include <assert.h>
#include <defines.h>
#include <math.h>
#include <point.h>
#include <stdio.h>
#include <system.h>
#include <sys/time.h>
#include <sys/types.h>

static aglo_boolean monitor_running;

static time_t start_time;
static time_t next_time;

static FILE *gpstream;			/* command stream to gnuplot */
static char gp_data_fname[] = "/tmp/glo.XXXXXX";
static char gp_replot_command[128];
static char *gpcommand = "gnuplot -geometry 600x700-1+1 >/dev/null 2>&1";
static char *gpconfig =
	"set terminal X11\n"
	"set data style linespoints\n"
	/* set format "" */
	"set offsets 0.1, 0.1, 0.1, 0.1\n"
	/* #set noxtics; set noytics */
	"set nokey\n"
	"set clip two\n";

static aglo_real t;

static void state_frame_points(aglo_graph graph, aglo_state_vector state,
                               aglo_point min_frame, aglo_point max_frame);

static void init_monitor(void) {
    /* only 2D for now */
    assert(AGLO_SPACE_ARITY == 2);

    gpstream = popen(gpcommand, "w");
    if (!gpstream) {
        perror("monitor: pipe open failed");
        exit(2);
    }
    if (EOF == fputs(gpconfig, gpstream)) {
        perror("monitor: first pipe write failed");
        exit(2);
    }
    mkstemp(gp_data_fname);
    if (!gp_data_fname) {
        perror("monitor: mktemp failed");
        exit(2);
    }
    sprintf(gp_replot_command, "plot \"%s\"\n", gp_data_fname);
    fflush(gpstream);

    monitor_running = true;
}

static void update_monitor(aglo_graph graph, aglo_state_vector state) {
    FILE *f = fopen(gp_data_fname, "w");
    char tbuf[80];
    aglo_point min_frame, max_frame, min_iso_frame, max_iso_frame;

    const aglo_real gp_plot_margin = 0.05;
    aglo_real gp_d;

    state_frame_points(graph, state, min_frame, max_frame);
    aglo_point_iso_frame(min_iso_frame, max_iso_frame, min_frame, max_frame);

    aglo_output_edge_state(f, graph, state);
    fclose(f);

    gp_d = gp_plot_margin * (max_iso_frame[0] - min_iso_frame[0]);

    sprintf(tbuf, "set xrange [ %f : %f ]\n",
            min_iso_frame[0] - gp_d, max_iso_frame[0] + gp_d);
    fputs(tbuf, gpstream);
    sprintf(tbuf, "set yrange [ %f : %f ]\n",
            min_iso_frame[1] - gp_d, max_iso_frame[1] + gp_d);
    fputs(tbuf, gpstream);

    sprintf(tbuf, "set title \"Time=%lu    Temp=%.3f\"\n",
            (unsigned long)(time(0) - start_time), t);
    fputs(tbuf, gpstream);

    fputs(gp_replot_command, gpstream);
    fflush(gpstream);
}

static void cleanup_monitor(void) {
    pclose(gpstream);
}

static void zero_state_vector(aglo_state_vector v, unsigned size) {
    unsigned i, j;

    for (i=0;i<size;i++)
        for (j=0;j<AGLO_SPACE_ARITY;j++) v[i][j] = 0;
}

static void random_placement(aglo_graph graph, aglo_state_vector state,
                             aglo_real placement_size) {
    unsigned i, j;

    for (i=0;i<graph->vertices;i++)
        for (j=0;j<AGLO_SPACE_ARITY;j++)
            state[i][j] = (drand48() - 0.5) * 2.0 * placement_size;
}

static void calculate_aesth_forces(aglo_graph graph, aglo_font font,
                                   aglo_state_vector state,
                                   aglo_state_vector gradient) {
    unsigned i, j;

    zero_state_vector(gradient, graph->vertices);

    for (; font->aesth_gradient; font++)
        if (font->enabled) {
            aglo_state_vector a_gradient;
            zero_state_vector(a_gradient, graph->vertices);
            font->aesth_gradient(graph, state, a_gradient);
            for (i=0;i<graph->vertices;i++)
                for (j=0;j<AGLO_SPACE_ARITY;j++)
                    gradient[i][j] += a_gradient[i][j] * font->weight;
        }
}

static void limit_displacement(aglo_real t, aglo_state_vector gradient,
                               unsigned size) {
    unsigned i, j;
    aglo_real sum_sqr = 0;
    aglo_real mag;

    for (i=0;i<size;i++)
        for (j=0;j<AGLO_SPACE_ARITY;j++)
            sum_sqr += sqr(gradient[i][j]);
    mag = sqrt(sum_sqr);
    if (mag > t) {
        mag = t / mag;
        for (i=0;i<size;i++)
            for (j=0;j<AGLO_SPACE_ARITY;j++)
                gradient[i][j] *= mag;
    }
}

static void make_move(aglo_state_vector state, aglo_state_vector gradient,
                      unsigned size) {
    unsigned i, j;

    for (i=0;i<size;i++)
        for (j=0;j<AGLO_SPACE_ARITY;j++)
            state[i][j] += gradient[i][j];
}

static void jitter(aglo_state_vector state, unsigned size) {
    unsigned long l = lrand48();
    unsigned i = l % size;
    unsigned j = l % AGLO_SPACE_ARITY;
    int sign = -1 + 2 * (l % 2); /* -1 or 1 */

    state[i][j] += sign * 0.00001;
}

static void do_aesth_setup(aglo_graph graph, aglo_font font) {
    for (; font->aesth_gradient; font++)
        if (font->enabled)
            font->aesth_setup(graph);
}

static void state_frame_points(aglo_graph graph, aglo_state_vector state,
                               aglo_point min_frame, aglo_point max_frame) {
    unsigned i, j;

    assert(graph->vertices > 0);

    aglo_point_assign(min_frame, state[0]),
    aglo_point_assign(max_frame, state[0]);
    for (i=0;i<graph->vertices;i++)
        for (j=0;j<AGLO_SPACE_ARITY;j++)
        {
            min_frame[j] = fmin(min_frame[j], state[i][j]),
            max_frame[j] = fmax(max_frame[j], state[i][j]);
        }
}


/* ---------------------------------------------------------------------- */

void aglo_startup(aglo_boolean do_monitor) {
    if (do_monitor) init_monitor();
}

void aglo_find_layout(aglo_graph graph, aglo_font font,
                      aglo_state_vector state,
                      aglo_find_params p, aglo_boolean random_start,
                      aglo_boolean do_monitor) {
    aglo_real k_lambda;
    unsigned long i;
    aglo_state_vector gradient;

    aglo_graph_sequence++;

    if (random_start) random_placement(graph, state, 1.0);

    assert(p->k_begin_temp > p->k_end_temp);
    assert(p->k_end_temp > 0);
    assert(p->k_iterations > 0);

    k_lambda = pow(p->k_begin_temp / p->k_end_temp, 1.0 / p->k_iterations);
    t = p->k_begin_temp;

    do_aesth_setup(graph, font);

    if (do_monitor) {
        start_time = time(0);
        next_time = start_time + p->mon_delay;
        update_monitor(graph, state);
    }

    for (i=1;i<=p->k_iterations;i++) {
        aglo_state_sequence++;
        calculate_aesth_forces(graph, font, state, gradient);
        limit_displacement(t, gradient, graph->vertices);
        make_move(state, gradient, graph->vertices);
        jitter(state, graph->vertices);
        t = t / k_lambda;
        if (do_monitor)
            if (time(0) >= next_time) {
                next_time += p->mon_delay;
                update_monitor(graph, state);
            }
    }

    if (do_monitor) update_monitor(graph, state);
}

void aglo_normalize_state(aglo_state_vector norm_state, aglo_graph graph,
                          aglo_state_vector state) {
    /* possibly norm_state == state */

    aglo_point min_frame, max_frame;
    aglo_point min_iso_frame, max_iso_frame;
    aglo_real iso_frame_size;
    unsigned i;

    state_frame_points(graph, state, min_frame, max_frame);
    aglo_point_iso_frame(min_iso_frame, max_iso_frame, min_frame, max_frame);
    iso_frame_size = max_iso_frame[0] - min_iso_frame[0];
    for (i=0;i<graph->vertices;i++) {
        aglo_point_sub(norm_state[i], state[i], min_iso_frame);
        if (iso_frame_size > 0)
            aglo_point_scalar_mult(norm_state[i], 1/iso_frame_size, norm_state[i]);
    }
}

void aglo_cleanup(aglo_boolean do_monitor) {
    if (do_monitor) cleanup_monitor();
}
